@@ -1,6 +1,7 @@
 """Convenience functions for storing and loading data."""
 import io
 import itertools
+import os
 import re
 from collections.abc import Mapping
 
@@ -100,6 +101,10 @@ def open(store=None, mode='a', **kwargs):
             raise PathNotFoundError(path)
 
 
+def _might_close(path):
+    return isinstance(path, (str, os.PathLike))
+
+
 def save_array(store, arr, **kwargs):
     """Convenience function to save a NumPy array to the local file system, following a
     similar API to the NumPy save() function.
@@ -131,7 +136,7 @@ def save_array(store, arr, **kwargs):
         array([   0,    1,    2, ..., 9997, 9998, 9999])
 
     """
-    may_need_closing = isinstance(store, str)
+    may_need_closing = _might_close(store)
     store = normalize_store_arg(store, clobber=True)
     try:
         _create_array(arr, store=store, overwrite=True, **kwargs)
@@ -202,7 +207,7 @@ def save_group(store, *args, **kwargs):
     if len(args) == 0 and len(kwargs) == 0:
         raise ValueError('at least one array must be provided')
     # handle polymorphic store arg
-    may_need_closing = isinstance(store, str)
+    may_need_closing = _might_close(store)
     store = normalize_store_arg(store, clobber=True)
     try:
         grp = _create_group(store, overwrite=True)
