@@ -53,3 +53,24 @@ try:
     have_fsspec = True
 except ImportError:  # pragma: no cover
     have_fsspec = False
+
+
+def abs_container():
+    from azure.core.exceptions import ResourceExistsError
+    import azure.storage.blob as asb
+
+    URL = "http://127.0.0.1:10000"
+    ACCOUNT_NAME = "devstoreaccount1"
+    KEY = "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw=="
+    CONN_STR = (
+        f"DefaultEndpointsProtocol=http;AccountName={ACCOUNT_NAME};"
+        f"AccountKey={KEY};BlobEndpoint={URL}/{ACCOUNT_NAME};"
+    )
+
+    blob_service_client = asb.BlobServiceClient.from_connection_string(CONN_STR)
+    try:
+        container_client = blob_service_client.create_container("test")
+    except ResourceExistsError:
+        container_client = blob_service_client.get_container_client("test")
+
+    return container_client
